private import
	std.path, std.file, std.c.stdio,
	dyayaml,
	runner, transform;

void print_usage()
{
	const char[] msg = 
		`yslt SOURCE.yaml [TRANSFORM.ysl]`\n
		`yslt -yamp SOURCE.yaml`\n;
	fprintf(stderr, `%.*s`, msg);
}

int main(char[][] args)
{
	void puts(char[] s){ printf("%.*s\n", std.string.replace(s, "\r\n", "\n")); }
	void err_puts(char[] s){ fprintf(stderr, "%.*s\n", s); }
	
	int result = 0;
	try{
		if(args.length <= 1){
			print_usage();
			result = 1;
		}else if(args[1] == `-yamp`){
			if(args.length <= 2){
				print_usage();
				result = 1;
			}else{
				YamlHolder source = yamlLoad(`YAMP`, cast(char[])std.file.read(args[2]));
				YamlLangRunner runner = new YamlLangRunner(null, source);
				runner.set_default_built_in_methods(&puts);
				runner.execute(`@main`);
			}
		}else{
			char[] source_file = args[1];
			char[] transform_file = ``;
			if(args.length <= 2){
				transform_file = std.path.addExt(source_file, `ysl`);
			}else{
				transform_file = args[2];
			}
			char[] source_text;
			try{
				source_text = cast(char[])std.file.read(source_file);
			}catch(FileException){
				source_file = std.path.defaultExt(source_file, `yaml`);
				try{
					source_text = cast(char[])std.file.read(source_file);
				}catch(FileException){
					source_file = std.path.addExt(source_file, `yml`);
					source_text = cast(char[])std.file.read(source_file);
				}
			}
			char[] transform_text = cast(char[])std.file.read(transform_file);
			YamlHolder source;
			YamlHolder transform;
			try{
				 source = yamlLoad(``, source_text);
			}catch(Error e){
				throw new Error(e.msg ~ ` on ` ~ source_file);
			}
			try{
				transform = yamlLoad(`YSLT`, transform_text);
			}catch(Error e){
				throw new Error(e.msg ~ ` on ` ~ transform_file);
			}
			yaml_transform(transform, source, &puts, &err_puts);
		}
	}catch(Exception e){
		fprintf(stderr, "error: %.*s\n", e.msg);
		result = 1;
	}
	return result;
}
