﻿private import
	std.string,
	dyayaml,
	runner;

class YamlTransformError : Error 
{
	this(char[] message){ super(message); }
}

const char[] ysl_head = `ysl-head`;
const char[] ysl_foot = `ysl-foot`;
const char[] ysl_list = `ysl-list`;
const char[] ysl_hash = `ysl-hash`;
const char[] ysl_item = `ysl-item`;
const char[] ysl_scalar = `ysl-scalar`;
const char[] ysl_any_list = `ysl-any-list`;
const char[] ysl_any_hash = `ysl-any-hash`;

const char[] ysl_key = `ysl-key`;
const char[] ysl_value = `ysl-value`;
const char[] ysl_indent = `ysl-indent`;
const char[] ysl_float = `ysl-float`;
const char[] ysl_skip = `ysl-skip`;
const char[] ysl_warning = `ysl-warning`;
const char[] ysl_error = `ysl-error`;

bit is_executable(YamlHolder node)
{
	if(node.is_scalar()){
		return true;
	}else if(node.is_hash()){
		foreach(char[] key, YamlHolder value; node.hash()){
			if(key.length > 0 && key[0] == '@'){
				return true;
			}
		}
	}
	return false;
}

void yaml_transform(YamlHolder rule_root, YamlHolder source_root, PutS puts, PutS err_puts)
{
	YamlLangRunner environment = new YamlLangRunner(null, rule_root);
	
	bit new_line = true;
	bit new_line_changed = false;
	char[] buffer;
	char[] indent;
	
	void yputs(char[] s)
	{
		//printf("(%.*s)", s);
		if(new_line){
			puts(indent ~ s);
		}else{
			if(buffer.length == 0){ buffer = indent; }
			buffer ~= s;
		}
	}
	
	void set_new_line(bit value)
	{
		if(value && !new_line && buffer.length > 0){
			puts(buffer);
			buffer = ``;
		}
		new_line = value;
		new_line_changed = true;
		//const char[][] b = ["F", "T"];
		//yputs(b[value]);
	}
	
	char[] current_key;
	char[] current_value;
	YamlHolder any_hash;
	YamlHolder any_list;
	
	environment.set_default_built_in_methods(&yputs);
	environment.built_in_delegates[ysl_key] = delegate YamlHolder(YamlLangContext context){
		yputs(current_key);
		return example_of_false;
	};
	environment.built_in_delegates[ysl_value] = delegate YamlHolder(YamlLangContext context){
		yputs(current_value);
		return example_of_false;
	};
	environment.built_in_delegates[ysl_indent] = delegate YamlHolder(YamlLangContext context){
		indent ~= \t;
		return example_of_false;
	};
	environment.built_in_delegates[ysl_float] = delegate YamlHolder(YamlLangContext context){
		//yputs(";");
		set_new_line(false);
		//yputs(";");
		return example_of_false;
	};
	environment.built_in_delegates[ysl_skip] = delegate YamlHolder(YamlLangContext context){
		return example_of_false;
	};
	environment.built_in_delegates[ysl_warning] = delegate YamlHolder(YamlLangContext context){
		try{
			char[] msg = context.get_variable(`msg`).string_value();
			err_puts(`**** warning **** : ` ~ msg);
		}catch(YamlLangError){
			err_puts(`**** warning ****`);
		}
		return example_of_false;
	};
	environment.built_in_delegates[ysl_error] = delegate YamlHolder(YamlLangContext context){
		try{
			char[] msg = context.get_variable(`msg`).string_value();
			err_puts(`**** error **** : ` ~ msg);
		}catch(YamlLangError){
			err_puts(`**** error ****`);
		}
		throw new YamlTransformError(`aborted by ysl-error.`);
		return example_of_false;
	};
	
	void execute(YamlHolder command)
	{
		if(command.is_scalar()){
			yputs(command.string_value());
		}else if(command.is_list()){
			bit old_new_line = new_line;
			set_new_line(false);
			new_line_changed = false;
			for(int i = 0; i < command.list().length; ++i){
				execute(command[i]);
			}
			if(!new_line_changed){ set_new_line(old_new_line); }
			//if(old_new_line){ yputs("*"); }
		}else{
			YamlLangMethodRunner runner = new YamlLangMethodRunner(environment, command);
			runner.variables[ysl_value] = new Yaml_Scalar(current_value);
			runner.execute();
		}
	}
	
	void process_node(YamlHolder rule, YamlHolder source)
	{
		void process_list(YamlHolder rule)
		{
			if(rule.is_list()){
				for(int i = 0; i < rule.list().length; ++i){
					process_list(rule[i]);
				}
			}else{
				assert(rule.is_hash());
				assert(source.is_list());
				char[] old_indent = indent;
				if(ysl_head in rule.hash()){
					execute(rule[ysl_head]);
				}
				if(!(ysl_item in rule.hash())){
					throw new YamlTransformError(ysl_item ~ ` is not existing in list.`);
				}
				YamlHolder item_rule = rule[ysl_item];
				int length = source.list().length;
				for(int i = 0; i < length; ++i){
					process_node(item_rule, source[i]);
				}
				if(ysl_foot in rule.hash()){
					execute(rule[ysl_foot]);
				}
				indent = old_indent;
			}
		}
		
		void process_hash(YamlHolder rule)
		{
			if(rule.is_list()){
				for(int i = 0; i < rule.list().length; ++i){
					process_hash(rule[i]);
				}
			}else{
				assert(rule.is_hash());
				assert(source.is_hash());
				char[] old_indent = indent;
				char[] old_key = current_key;
				if(ysl_head in rule.hash()){
					execute(rule[ysl_head]);
				}
				YamlHolder default_rule;
				if(ysl_item in rule.hash()){
					default_rule = rule[ysl_item];
				}
				Yaml_Hash source_hash = cast(Yaml_Hash)source;
				assert(source_hash !== null);
				assert(source_hash.keys_valid());
				int length = source_hash.keys().length;
				for(int i = 0; i < length; ++i){
					current_key = source_hash.keys()[i];
					YamlHolder elem = source_hash[current_key];
					if(current_key in rule.hash()){
						process_node(rule[current_key], elem);
					}else{
						if(default_rule === null){
							throw new YamlTransformError(ysl_item ~ ` is not existing in hash.`);
						}
						process_node(default_rule, elem);
					}
				}
				if(ysl_foot in rule.hash()){
					execute(rule[ysl_foot]);
				}
				current_key = old_key;
				indent = old_indent;
			}
		}
		
		if(rule.is_scalar()){
			execute(rule);
		}else if(rule.is_list()){
			//yputs("(");
			bit old_new_line = new_line;
			bit first = true;
			new_line_changed = false;
			for(int i = 0; i < rule.list().length; ++i){
				YamlHolder r = rule[i];
				if(is_executable(r)){
					if(first){ set_new_line(false); new_line_changed = false; }
					if(source.is_scalar()){
						char[] old_value = current_value;
						current_value = source.string_value();
						execute(r);
						current_value = old_value;
					}else{
						execute(r);
					}
				}else{
					process_node(r, source);
				}
				first = false;
			}
			if(!new_line_changed){ set_new_line(old_new_line); }
			//yputs(")");
		}else{
			assert(rule.is_hash());
			YamlHolder old_any_hash = any_hash;
			YamlHolder old_any_list = any_list;
			if(ysl_any_hash in rule.hash()){
				any_hash = rule[ysl_any_hash];
			}
			if(ysl_any_list in rule.hash()){
				any_list = rule[ysl_any_list];
			}
			if(ysl_head in rule.hash()){
				execute(rule[ysl_head]);
			}
			if(source.is_list()){
				//yputs("[");
				bit old_new_line = new_line;
				if(is_executable(rule)){
					execute(rule);
				}else{
					YamlHolder r = any_list;
					if(ysl_list in rule.hash()){
						r = rule[ysl_list];
					}
					if(r === null){
						throw new YamlTransformError(ysl_list ~ ` is not existing.`);
					}
					if(is_executable(r)){
						execute(r);
					}else{
						process_list(r);
					}
				}
				set_new_line(old_new_line);
				//if(old_new_line){ yputs("*"); }
				//yputs("]");
			}else if(source.is_hash()){
				//set_new_line(true);
				//yputs("{");
				bit old_new_line = new_line;
				if(is_executable(rule)){
					execute(rule);
				}else{
					assert(rule.is_hash());
					YamlHolder r = any_hash;
					if(ysl_hash in rule.hash()){
						r = rule[ysl_hash];
					}
					if(r === null){
						//err_puts(format("%d", source.hash().length));
						//err_puts(format("%d", rule.hash().length));
						throw new YamlTransformError(ysl_hash ~ ` is not existing.`);
					}
					if(is_executable(r)){
						execute(r);
					}else{
						process_hash(r);
					}
				}
				set_new_line(old_new_line);
				//yputs("}");
			}else{
				assert(source.is_scalar());
				char[] old_value = current_value;
				current_value = source.string_value();
				if(ysl_scalar in rule.hash()){
					process_node(rule[ysl_scalar], source);
				}else{
					execute(rule);
				}
				current_value = old_value;
			}
			if(ysl_foot in rule.hash()){
				execute(rule[ysl_foot]);
			}
			any_hash = old_any_hash;
			any_list = old_any_list;
		}
	}
	
	process_node(rule_root, source_root);
}
